/*
 * Main.java
 *
 * Created on 12 novembre 2005, 8.39
 */

import java.lang.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;

/**
 * Classe principale contenente l'algoritmo risolutivo.
 *
 * @author Daniele Baracchi
 */
public class Main {
    
    // Percorsi prefissati delle varie linee (in un solo senso)
    public final static char [][] percorsi = {
        {'A', 'B', 'C', 'D'},
        {'E', 'F', 'G', 'C', 'H', 'I', 'L'},
        {'M', 'N', 'O', 'C', 'P', 'Q', 'R'},
        {'D', 'C', 'P', 'T', 'S'}
    };
    
    // Funzione usata per "ribaltare" l'array delle fermate e formare i percorsi inversi
    private static char [] reverse(char [] v) {
        char [] r = new char[v.length];
        for (int i = 0; i < v.length; i++)
            r[r.length-1-i] = v[i];
        return r;
    }
    
    // Lista degli autisti disponibili
    protected AutistiList autisti;
    // Lista delle richieste effettuate
    protected RichiesteList richieste;
    // Lista delle linee (incluse quelle dei percorsi inversi)
    protected ArrayList<Linea> linee = new ArrayList<Linea>(8);
    // Array di insiemi. L'i-esimo insieme contiene le richieste della i-esima linea
    protected HashSet<Richiesta> ricl[] = new HashSet[8];
    // Insieme di utenti insoddisfatti (sperando che non si debba mai riempire)
    protected HashSet<Richiesta> insod = new HashSet<Richiesta>();
    // Lista di corse create a partire dai dati iniziali
    protected LinkedList<Corsa> corse = new LinkedList<Corsa>();
    // Insieme degli utenti soddisfatti (sperando che li contenga tutti)
    protected HashSet<Risposta> sodd = new HashSet<Risposta>();
    
    /**
     * Costruttore della classe.
     * Essenzialmente, svolge lui il "lavoro sporco".
     * @param autisti La lista autisti letta da file
     * @param richieste La lista delle richieste letta da file
     */
    public Main(AutistiList autisti, RichiesteList richieste) {
        // Aggiunge tutte le linee alla lista
        linee.add(new Linea(percorsi[0]));
        linee.add(new Linea(percorsi[1]));
        linee.add(new Linea(percorsi[2]));
        linee.add(new Linea(percorsi[3]));
        // Aggiunge le linee col percorso inverso
        linee.add(new Linea(reverse(percorsi[0])));
        linee.add(new Linea(reverse(percorsi[1])));
        linee.add(new Linea(reverse(percorsi[2])));
        linee.add(new Linea(reverse(percorsi[3])));
        this.autisti = autisti;
        this.richieste = richieste;
        
        // Genera gli insiemi di richieste
        for (int i = 0; i < 8; i++)
            ricl[i] = new HashSet<Richiesta>();
        
        // Popola ogni insieme con le richieste di competenza
        for (Richiesta r : richieste)
            ricl[r.linea].add(r);
        
        // Variabile usata unicamente per determinare la fine dell'algoritmo
        boolean nxt = true;
        do {
            for (int i = 0; i < 8; i++) {
                // Se l'insieme delle richieste da elaborare per questa linea e' vuoto, la saltiamo
                if (ricl[i].isEmpty())
                    continue;
                
                /* Crea un array enorme per contenere il numero di richieste all'interno di tutti i possibili
                 * sottointervalli di 30 minuti contenuti nella giornata.
                 * Gli intervalli sono di 30 minuti dato che per ogni richiesta e' ammesso uno scarto in negativo
                 * o in positivo di 15 minuti.
                 */
                int count[] = new int[12*11];
                for (int k = 0; k < count.length; k++)
                    count[k] = 0;
                
                // Crea una lista di richieste altrettanto enorme per contenere le eventuali richieste
                LinkedList<Richiesta> inter[] = new LinkedList[12*11];
                
                for (int h = 0; h < 12*11; h++)
                    inter[h] = new LinkedList<Richiesta>();
                
                // Scorre tutti i sottointervalli, incrementando il contatore e aggiungendo alla lista
                Orario ora = new Orario(8, 00);
                for (int j = 0; j < 12*11; j++, ora = ora.add(5)) {
                    for (Richiesta r : ricl[i]) {
                        int m = r.partenza.difference(
                                ora.add(5*linee.get(i).getOrdinal(r.origine))
                                );
                        if (m <= 15) {
                            count[j]++;
                            inter[j].add(r);
                        }
                    }
                }
                
                // Seleziona il sottointervallo con il massimo numero di richieste
                int max = 0;
                for (int j = 1; j < 12*11; j++) {
                    if (count[max] < count[j])
                        max = j;
                }
                
                // Calcola il numero di taxi necessari (basandosi sul fatto che 1 taxi == 3 persone)
                int needed = (count[max] / 3 > 0 ? count[max] / 3 +
                        (count[max] % 3 > 0 ? 1 : 0): 1);
                
                // Itera n volte, con n == numero di taxi necessari
                for (int w = 0; w < needed; w++) {
                    // Cerca un autista disponibile per la fascia oraria indicata
                    Orario partenza = new Orario(8,00).add(5*max);
                    int aut = findAvailable(partenza,
                            partenza.add(5*(linee.get(i).fermate.size()-1)));
                    // Se l'autista e' disponibile crea le corse, le assegna e sposta gli utenti tra i soddisfatti
                    if (aut > -1) {
                        Corsa gen = new Corsa(aut, i, partenza, linee.get(i).fermate.size());
                        corse.add(gen);
                        // Tratta in modo diverso il caso in cui il taxi sia pieno, o occupato solo in parte
                        if (count[max] / 3 == 0)
                            for (Richiesta rc : inter[max]) {
                                sodd.add(new Risposta(rc, gen));
                                ricl[i].remove(rc);
                                count[max]--;
                            }
                        else
                            for (int g = 0; g < 3; g++) {
                                sodd.add(new Risposta(inter[max].get(0), gen));
                                ricl[i].remove(inter[max].remove(0));
                                count[max]--;
                            }
                    // Se non ci sono autisti disponibili, sposta gli utenti tra gli insoddisfatti
                    } else {
                        for (Richiesta rc : inter[max]) {
                            insod.add(rc);
                            ricl[i].remove(rc);
                        }
                    }
                }
            }
            
            // Controlla se occorre terminare il ciclo guardando se ci sono ancora richieste da elaborare
            nxt = false;
            for (int i = 0; i < 8; i++)
                if (!ricl[i].isEmpty())
                    nxt = true;
            
        } while (nxt);
        
    }
    
    /** Controlla se c'e' un autista libero per l'orario richiesto.
      * @return L'indice dell'autista nella lista, o -1 se non ce ne sono disponibili
      */
    protected int findAvailable(Orario inizio, Orario fine) {
        // Itera su tutti gli autisti alla ricerca del primo disponibile
        for (int i = 0; i < autisti.length(); i++) {
            boolean ok = true;
            
            /* Itera su tutte le corse (controllando che siano effettuate proprio da quell'autista)
             * per vedere che non ci siano problemi di orario. Considera anche il tempo di riposo di 30 minuti
             */
            for (Corsa c : corse) {
                if (c.autista != i)
                    continue;
                
                if (c.arrivo.compareTo(inizio) < 0 &&
                        c.arrivo.difference(inizio) >= 30)
                    continue;
                
                if (c.partenza.compareTo(fine) > 0 &&
                        c.partenza.difference(fine) >= 30)
                    continue;
                
                // Se siamo arrivati fin qui significa che l'autista ha impegni e non e' disponibile
                ok = false;
                break;
            }
            
            // Se abbiamo trovato un autista disponibile, allora ritorniamo il suo indice
            if (ok)
                return i;
        }
        
        // Evidentemente, nessun autista era disponibile
        return -1;
    }
    
    /**
     * Il metodo main legge da file, richiama l'algoritmo, e scrive i risultati.
     * @param args Non usato
     */
    public static void main(String[] args) {
        // Crea le liste a partire dai file dati
        final AutistiList autisti = new AutistiList("AUTISTI.TXT");
        final RichiesteList richieste = new RichiesteList("RICHIESTE.TXT");
        
        // Genera la soluzione
        Main soluzione = new Main(autisti, richieste);
        
        /* Scrive i risultati su file secondo le specifiche date.
         * Da notare la conversione effettuata sul numero linea in base all'indice nell'array.
         * [0] -> 11 ; [4] -> 12
         * [1] -> 21 ; [5] -> 22
         * [2] -> 31 ; [6] -> 32
         * [3] -> 41 ; [7] -> 42
         * Che in Java si traduce in (linea < 4 ? (linea + 1) + "1" : (linea - 3) + "2").
         */
        try {
            OutputStreamWriter corse = new OutputStreamWriter(new FileOutputStream("OUTPUT_CORSE.TXT"));
            
            int n = 0;
            
            Collections.sort(soluzione.corse, new Comparator<Corsa>() {
                public int compare(Corsa c1, Corsa c2) {
                    return autisti.get(c1.autista).nome.compareTo(autisti.get(c2.autista).nome);
                }
            });
            
            for (Corsa c : soluzione.corse) {
                corse.write(autisti.get(c.autista).nome + "\t" +
                        autisti.get(c.autista).telefono + "\t" +
                        (c.linea < 4 ? (c.linea + 1) + "1" : (c.linea - 3) + "2" ) + "\t" +
                        n++ + "\t" + c.partenza + "\t" + c.arrivo + "\n");
            }
            
            corse.close();
            
            OutputStreamWriter sodd = new OutputStreamWriter(new FileOutputStream("OUTPUT_UTENTI_SODD.TXT"));
            
            ArrayList<Risposta> ar = new ArrayList<Risposta>(soluzione.sodd);
            
            /* Per ordinare la lista delle risposte viene creata al volo una inner class
             * che compara in base al nome del richiedente
             */
            Collections.sort(ar, new Comparator<Risposta>() {
                public int compare(Risposta r1, Risposta r2) {
                    return r1.originale.nome.compareTo(r2.originale.nome);
                }
            });
            
            for (Risposta r : ar) {
                Richiesta ric = r.originale;
                int nfp = soluzione.linee.get(ric.linea).getOrdinal(ric.origine),
                        nfa = soluzione.linee.get(ric.linea).getOrdinal(ric.destinazione);
                sodd.write(ric.nome + "\t" + ric.telefono + "\t" +
                        (ric.linea < 4 ? (ric.linea + 1) + "1" :
                            (ric.linea - 3) + "2") + "\t" +
                        soluzione.corse.indexOf(r.generata) + "\t" +
                        ric.origine + "\t" + ric.partenza + "\t" +
                        r.generata.partenza.add(nfp*5) + "\t" + ric.destinazione + "\t" +
                        r.generata.partenza.add(nfa*5) + "\n");
            }
            
            sodd.close();
            
            OutputStreamWriter insodd = new OutputStreamWriter(new FileOutputStream("OUTPUT_UTENTI_NON_SODD.TXT"));
            
            for (Richiesta r : soluzione.insod) {
                insodd.write(r.nome + "\t" + r.telefono + "\t" +
                        (r.linea < 4 ? (r.linea + 1) + "1" :
                            (r.linea - 3) + "2") + "\t" +
                        r.origine + "\t" + r.destinazione + "\t" +
                        r.partenza + "\n");
            }
            
            insodd.close();
        } catch (IOException ex) {
            System.out.println("Errore di scrittura su file.");
        }
        
        System.exit(0);
    }
    
}
